from typing import Iterator, List
import pathlib

import cq.checker


class InitChecker(cq.checker.Checker):
	NAME = 'init_checker'
	DESCRIPTION = 'Checks for `__init__.py` files in subdirectories containing `.py` files.'
	HELP_LINE = 'create `__init__.py` file in the particular directory'
	MESSAGE = 'Directory is missing `__init__.py` file.'

	def run(self, modules: List[str]) -> cq.checker.CheckerResult:
		missing_inits = list(
			result
			for module in modules
			if pathlib.Path(module).is_dir()
			for result in find_missing_inits(pathlib.Path(module))
		)
		output = [
			cq.checker.ResultLine(file = str(result), line = None, message = self.MESSAGE, is_error = False)
			for result in missing_inits
		]
		if output:
			header = '\n\nTo create missing __init__.py files, run:\n'
			cmd = f'touch {" ".join(str(path / "__init__.py") for path in missing_inits)}'
			output.append(
				cq.checker.ResultLine(file = None, line = None, message = f'{header}{cmd}', is_error = False)
			)
		return cq.checker.CheckerResult(
			checker_name = self.NAME,
			help_line = self.HELP_LINE,
			return_code = 0,
			output = output,
		)


def find_missing_inits(directory: pathlib.Path) -> Iterator[pathlib.Path]:
	'''
	Recursively yield subdirectories that contain *.py files but are missing __init__.py.
	Ignores hidden directories.
	'''
	for subdir in directory.iterdir():
		if subdir.is_dir() and not subdir.name.startswith('.'):
			if any(subdir.glob('*.py')) and not (subdir / '__init__.py').exists():
				yield subdir
			yield from find_missing_inits(subdir)



from typing import IO, List, Optional, Tuple
import configparser
import os.path
import re
import subprocess
import tempfile

import cq.checker
import cq.utils


class PylintChecker(cq.checker.Checker):
	NAME = 'pylint'
	DESCRIPTION = 'Runs pylint on given path.'
	HELP_LINE = 'use `# pylint: disable=<violation name>` for disabling line check. For a list of violations, see `pylint --list-msgs`'

	@staticmethod
	def _get_pylint_rc_file(default_rcfile_path: str) -> Tuple[str, Optional[IO[str]]]:
		'''
		Try to find .pylintr rc in $PWD and if found merge it with the default one
		for usage with pylint checker
		'''
		override_rcfile_path = os.path.join(os.getcwd(), '.pylintrc')
		if os.path.exists(override_rcfile_path):
			config = configparser.ConfigParser()
			config.read(default_rcfile_path)
			config.read(override_rcfile_path)
			tmp_file = tempfile.NamedTemporaryFile(mode = 'w+')
			# a bug of tempfile typeshed https://github.com/python/typeshed/issues/434
			config.write(tmp_file)
			tmp_file.flush()
			return tmp_file.name, tmp_file

		return default_rcfile_path, None

	@staticmethod
	def _setup_pylint() -> Tuple[List[str], Optional[IO[str]]]:
		'''
		Setup pylint.

		:return path to the configuration file; if not None, this is file handle for temporarily created conf. file
		'''
		# --rcfile must be always last!
		pylint_options = ['--reports=no', '--score=no', '--rcfile']
		rcfile_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'pylintrc')
		effective_rcfile_path: str
		tmp_file: Optional[IO[str]]
		effective_rcfile_path, tmp_file = PylintChecker._get_pylint_rc_file(rcfile_path)
		pylint_options.append(effective_rcfile_path)

		return pylint_options, tmp_file

	@staticmethod
	def _cleanse_packages(package_list: List[str]) -> List[str]:
		'''
		Remove packages from list, that are childs of each other.
		'''
		modules: List[str] = []
		for module in package_list:
			for added_module in modules:
				# we should not add module that is subpath of already added module
				# except ending with .py (so tests without __init__.py are also included)
				if module.startswith(added_module) and module != added_module and not module.endswith('.py'):
					break
			else:
				modules.append(module)
		return modules

	@staticmethod
	def _process_line(line: str) -> Tuple[Optional[str], Optional[int], str, bool]:
		'''
		Process single output line.

		:param line: Line to parse.
		:return: Parsed output in Tuple[file name, line number, message, is_error]
		'''
		if not line:
			return None, None, '', False

		if line.startswith('************* '):
			return None, None, '', False

		try:
			file, linenum, *msg = line.split(':')
		except ValueError:
			return None, None, line, False

		if not msg:
			return None, None, line, False

		message = ':'.join(msg).lstrip()

		is_error = True
		if re.search(r'\[[CRI]{1}\d{4}\([-\w]+\).*\]', message):
			is_error = False

		try:
			return file, int(linenum), message, is_error
		except ValueError:  # int conversion failed
			return None, None, line, False

	def run(self, modules: List[str]) -> cq.checker.CheckerResult:
		packages_to_cleanse = list(
			filter(lambda module: not module.endswith('fabfile.py'), cq.utils.get_path_generator(modules))
		)

		# remove packages, that are childs of other
		pylint_modules_to_check = PylintChecker._cleanse_packages(packages_to_cleanse)

		pylint_options, tmp_file = PylintChecker._setup_pylint()

		# redirect stderr to /dev/null because it contains line: Using config file .pylintrc
		stdout, pylint_return_code = cq.utils.run_external_checker(
			'pylint',
			pylint_options + pylint_modules_to_check,
			stderr = subprocess.DEVNULL,
		)

		if tmp_file:
			tmp_file.close()

		# Pylint should leave with following status code:
		# 0 if everything went fine
		# 1 if a fatal message was issued
		# 2 if an error message was issued
		# 4 if a warning message was issued
		# 8 if a refactor message was issued
		# 16 if a convention message was issued
		# 32 on usage error
		# status 1 to 16 will be bit-ORed so you can know which different
		# categories has been issued by analysing pylint output status code

		return_code = 0
		if pylint_return_code & 4 or pylint_return_code & 2 or pylint_return_code & 1:
			# Warning, error or fatal message occured
			return_code = 1
		elif pylint_return_code == 32:
			# PyLint usage error
			return_code = 2

		output: List[cq.checker.ResultLine] = []

		for line in stdout.split('\n'):
			file, linenum, message, is_error = PylintChecker._process_line(line)

			if not message:
				continue

			output.append(cq.checker.ResultLine(file = file, line = linenum, message = message, is_error = is_error))

		return cq.checker.CheckerResult(
			checker_name = self.NAME,
			help_line = self.HELP_LINE,
			return_code = return_code,
			output = output,
		)



from typing import Any, Callable, Iterable, Iterator, List, Match, Pattern, Tuple, cast
import pathlib
import re

import toolz

import cq.checker
import cq.utils


MessageGenerator = Callable[[str, Match], str]
Condition = Callable[[str, Match], bool]
Rule = Tuple[Pattern, MessageGenerator, Condition]


def compile_rule_specs(*rule_specs: Tuple[str, int, MessageGenerator, Condition]) -> List[Rule]:
	return [
		(re.compile(pattern, flags = flags), message_generator, condition)
		for (pattern, flags, message_generator, condition) in rule_specs
	]


def check_file_lines(preprocessor: Callable[[str], str], rules: List[Rule], file: pathlib.Path) -> str:
	output: List[str] = []
	with file.open() as f:
		content = f.read()
		preprocessed = preprocessor(content)
		for i, (line, original_line) in enumerate(zip(preprocessed.split('\n'), content.split('\n'))):
			for complaint in check_line(line, original_line, rules):
				output.append('{file}:{line_no}: {complaint}'.format(file = file, line_no = i + 1, complaint = complaint))
	return '\n'.join(output)


def check_line(line: str, original_line: str, rules: Iterable[Rule]) -> Iterator[str]:
	for pattern, message_generator, condition in rules:
		match = pattern.search(line)
		if match is not None:
			if condition(original_line, match):
				yield message_generator(line, match)


def _const(msg: str) -> Callable[[Any, Any], str]:
	return lambda _, __: msg


class GrammarNaziChecker(cq.checker.Checker):
	NAME = 'grammar_nazi'
	DESCRIPTION = 'Checks grammar rules in files'
	HELP_LINE = (
		'use `# grammar_chcker: disable=<violation name>` for disabling line check. '
		'Violations: 3rd-person-singular, happend, it-self, insturment, splitted.'  # grammar_checker: disable = happend, grammar_checker: disable = insturment, grammar_checker: disable = splitted
	)
	RULES = compile_rule_specs(
		(
			r'(does|may|might|can) not exists',
			re.IGNORECASE,
			lambda _, match: (
				'Use "{verb} not exist", not "{verb} not exists". '
				'The "{verb}" verb already is in third person singular.'
			).format(verb = match.group(1)),
			lambda original_line, _: re.search(r'grammar_checker:\s*disable\s*=\s*3rd-person-singular', original_line)
			is None,
		),
		(
			r'happend',  # grammar_checker: disable = happend
			re.IGNORECASE,
			_const(
				'"happend" is not a word. You either mean "happen" (present tense) or "happened" (past tense).'  # grammar_checker: disable = happend
			),
			lambda original_line, _: re.search(
				r'grammar_checker:\s*disable\s*=\s*happend',  # grammar_checker: disable = happend
				original_line,
			)
			is None,
		),
		(
			r'\bit self',
			re.IGNORECASE,
			_const('"itself" is one word.'),
			lambda original_line, _: re.search(r'grammar_checker:\s*disable\s*=\s*it-self', original_line) is None,
		),
		(
			r'insturment',  # grammar_checker: disable = insturment
			re.IGNORECASE,
			_const('"insturment", did you mean "instrument"'),  # grammar_checker: disable = insturment
			lambda original_line, _: re.search(
				r'grammar_checker:\s*disable\s*=\s*insturment',  # grammar_checker: disable = insturment
				original_line,
			)
			is None,
		),
		(
			r'splitted',  # grammar_checker: disable = splitted
			re.IGNORECASE,
			_const('The past tense of "to split" is "split", not "splitted".'),  # grammar_checker: disable = splitted
			lambda original_line, _: re.search(
				r'grammar_checker:\s*disable\s*=\s*splitted',  # grammar_checker: disable = splitted
				original_line,
			)
			is None,
		),
	)

	def run(self, modules: List[str]) -> cq.checker.CheckerResult:
		output: List[cq.checker.ResultLine] = []

		for file in cq.utils.get_all_python_files(modules):
			result = check_file_lines(preprocessor = lambda x: x, rules = self.RULES, file = pathlib.Path(file))

			for line in result.split('\n'):
				if not line:
					continue

				try:
					file, linenum, *msg = line.split(':')
				except ValueError:  # fallback solution, so it won't crash cq
					output.append(cq.checker.ResultLine(file = None, line = None, message = line, is_error = True))

					continue

				output.append(
					cq.checker.ResultLine(
						file = file,
						line = int(linenum),
						message = ':'.join(msg).lstrip(),
						is_error = True,
					)
				)

		return cq.checker.CheckerResult(
			checker_name = self.NAME,
			help_line = self.HELP_LINE,
			return_code = 0 if not output else 1,
			output = output,
		)


# This regexp is not perfect but we don't want to go crazy from this, do we.
# https://stackoverflow.com/a/14366904
_STRING_PATTERN = re.compile(r'''(\"\"\"|\'\'\'|\"|\')(((?<!\\)\\\1|.)*?)\1''', flags = re.DOTALL)
_COMMENT_PATTERN = re.compile(r'\#.*')


def clobber_string_literals(source: str) -> str:
	'''
	Hide string literal content by overwriting it with '.' characters.
	'''
	return _STRING_PATTERN.sub(_string_literal_cloberrer, source)


def _string_literal_cloberrer(match: Match) -> str:
	delimiter = match.group(1)
	content = match.group(2)
	clobbered = re.sub(r'\S', '.', content)
	# mypy considers the variables as Any
	return cast(str, delimiter + clobbered + delimiter)


def clobber_comments(source: str) -> str:
	'''
	Hide comment content by overwriting it with '#' characters.
	Make sure we don't identify '#' inside string literals as comments.

	>>> clobber_comments('x = 1 # 123')
	'x = 1 #####'
	>>> clobber_comments('x = "#"')
	'x = "#"'
	>>> clobber_comments('x = "#" + "x # y"  # comment')
	'x = "#" + "x # y"  #########'
	'''
	lines = []
	for line in source.split('\n'):
		index = line.find('#')
		while index > -1:
			for string_literal in _STRING_PATTERN.finditer(line):
				start, stop = string_literal.span()
				if start < index < stop:
					# The '#' character is inside a string literal. Don't clobber it and go looking for the next '#' on this line.
					index = line.find('#', index + 1)
					break
			else:
				# The '#' character is not inside any string literal => clobber the rest of the line.
				comment = line[index:]
				line = line[:index] + '#' * len(comment)
				break

		lines.append(line)

	return '\n'.join(lines)


def _match_surrounding(line: str, match: Match) -> str:
	return line[max(0, match.start() - 20) : min(len(line), match.end() + 20)]


class DumbStyleChecker(cq.checker.Checker):
	NAME = 'dumb_style_checker'
	DESCRIPTION = 'Checks some basic Pyhon mistakes'

	# '=' must come after '==' so that '==' matches first!
	_EQUALS_OPERATORS = ('==', '!=', '<=', '>=', '-=', '+=', '*=', '/=', '|=', '&=', '^=', ':=', '=')

	RULES = compile_rule_specs(
		(
			'((?<![ {other_chars}\t])({operators}))|(({operators})(?=[^ {other_chars}]))'.format(
				other_chars = ''.join({re.escape(op[0]) for op in _EQUALS_OPERATORS}),
				operators = '|'.join(re.escape(op) for op in _EQUALS_OPERATORS),
			),
			0,
			lambda line, match: 'Put exactly one space before and after `{}`  [... {} ...].'.format(
				match.group(0),
				_match_surrounding(line, match),
			),
			lambda *_: True,
		),
		(
			'^.*([a-z]+(2|4)[a-z]+)',
			re.IGNORECASE,
			lambda line, match: 'Do not spell `{}` as the digit `{}`. [... {} ...]'.format(
				'to' if match.group(2) == '2' else 'for',
				match.group(2),
				_match_surrounding(line, match),
			),
			lambda original_line, _: re.search(r'dumb_style_checker:\s*disable\s*=\s*word_as_a_digit', original_line)
			is None,
		),
		(
			r'\bprint\b',
			0,
			lambda line, match: (
				'It seems that you have forgotten about a print statement. '
				'If this is intentional, disable this check by # dumb_style_checker:disable = print-statement'
			),
			lambda original_line, _: re.search(r'dumb_style_checker:\s*disable\s*=\s*print-statement', original_line)
			is None,
		),
	)

	def run(self, modules: List[str]) -> cq.checker.CheckerResult:
		output: List[cq.checker.ResultLine] = []

		for file in cq.utils.get_all_python_files(modules):
			result = check_file_lines(
				preprocessor = toolz.compose(clobber_string_literals, clobber_comments),
				rules = self.RULES,
				file = pathlib.Path(file),
			)

			for line in result.split('\n'):
				if not line:
					continue

				try:
					file, linenum, *msg = line.split(':')
				except ValueError:
					output.append(cq.checker.ResultLine(file = None, line = None, message = line, is_error = True))

					continue

				output.append(
					cq.checker.ResultLine(
						file = file,
						line = int(linenum),
						message = ':'.join(msg).lstrip(),
						is_error = True,
					)
				)

		return cq.checker.CheckerResult(
			checker_name = self.NAME,
			help_line = None,
			return_code = 0 if not output else 1,
			output = output,
		)



from typing import Any, Callable, Iterable, List, NamedTuple, Optional, Tuple
import concurrent.futures
import logging
import sys
import time

import click

import cq.checker
import cq.checkers.bellybutton
import cq.checkers.git
import cq.checkers.init
import cq.checkers.isort
import cq.checkers.mypy
import cq.checkers.orange
import cq.checkers.pyflakes
import cq.checkers.pylint
import cq.checkers.regex
import cq.checkers.requirements
import cq.checkers.setup
import cq.fixer
import cq.fixers.isort_fixer
import cq.fixers.orange


checkers: List[cq.checker.Checker] = [
	cq.checkers.pylint.PylintChecker(),
	cq.checkers.mypy.MypyChecker(),
	cq.checkers.bellybutton.BellybuttonChecker(),
	cq.checkers.pyflakes.PyflakesExtChecker(),
	cq.checkers.regex.GrammarNaziChecker(),
	cq.checkers.regex.DumbStyleChecker(),
	cq.checkers.requirements.RequirementsSetupCompatiblityChecker(),
	cq.checkers.requirements.RequirementsvalidatorChecker(),
	cq.checkers.setup.SetupChecker(),
	cq.checkers.git.BranchNameChecker(),
	cq.checkers.init.InitChecker(),
	cq.checkers.orange.OrangeChecker(),
	cq.checkers.isort.IsortChecker(),
]
fixers: List[cq.fixer.Fixer] = [
	cq.fixers.isort_fixer.IsortFixer(),
	cq.fixers.orange.OrangeFixer(),
]


CHECK_MARK = '\u2713'
FAIL_X = '\u2717'


def help_callback(ctx: click.Context, _param: Any, value: Any) -> None:
	if not value:
		return

	formatter = ctx.make_formatter()
	ctx.command.format_usage(ctx, formatter)
	formatter.write('\nAvailable Checkers:\n')
	for checker in checkers:
		formatter.write(' - {}: {}\n'.format(checker.NAME, checker.DESCRIPTION))
	formatter.write('\nAvailable Fixers:\n')
	for fixer in fixers:
		formatter.write(' - {}: {}\n'.format(fixer.NAME, fixer.DESCRIPTION))
	ctx.command.format_help_text(ctx, formatter)
	ctx.command.format_options(ctx, formatter)
	ctx.command.format_epilog(ctx, formatter)
	click.echo(formatter.getvalue())
	ctx.exit()


class ModuleTime(NamedTuple):
	name: str
	time: float


module_time_elapsed: List[ModuleTime] = []


def measure_time(
	logger: logging.Logger,
	module_name: str,
	runner: Callable[..., cq.checker.CheckerResult],
	*args: Any,
	**kwargs: Any,
) -> Callable[[], cq.checker.CheckerResult]:
	global module_time_elapsed  # pylint: disable = global-statement

	def wrapper() -> cq.checker.CheckerResult:
		logger.debug('starting module %s', module_name)
		time_started = time.monotonic()

		result = runner(*args, **kwargs)

		time_elapsed = time.monotonic() - time_started
		logger.debug('module %s ended; time elapsed %.3f s', module_name, time_elapsed)
		module_time_elapsed.append(ModuleTime(name = module_name, time = time_elapsed))

		return result

	return wrapper


def run_checkers(
	list_modules: List[str],
	disable: Iterable[str],
	logger: logging.Logger,
	debug: bool,
	show_warnings: bool,
	threads: Optional[int],
) -> int:
	if debug:
		logging.debug('cq started in debug mode')

	cq_started = time.monotonic()

	if debug:
		logging.debug('using %s threads', threads or 'maximum available')

	exit_code = 0
	with concurrent.futures.ThreadPoolExecutor(max_workers = threads) as executor:
		futures = []

		for checker in checkers:
			if checker.NAME in disable:
				if debug:
					logging.debug('module %s disabled -> skipping', checker.NAME)

				continue

			if debug:
				runner = measure_time(logger, checker.NAME, checker.run, list_modules)
				futures.append(executor.submit(runner))
			else:
				futures.append(executor.submit(checker.run, list_modules))

		for future in concurrent.futures.as_completed(futures):
			result = future.result()

			exit_code = max(result.return_code, exit_code)

			if result.output:
				header_text = result.checker_name

				if result.help_line:
					header_text += f'\nHint: {result.help_line}'

				# with `show_warnings` disabled, only print header if there was an error
				if show_warnings or result.return_code:
					click.secho(header_text, bg = 'white', fg = 'black', bold = True, underline = True)

				for line_of_output in result.output:
					filename = ''
					if line_of_output.file is not None:
						filename += f'{line_of_output.file}:'

					if line_of_output.line is not None:
						filename += f'{line_of_output.line}:'

					if show_warnings or line_of_output.is_error:
						if filename:
							click.secho(f'{filename} {line_of_output.message}', bold = line_of_output.is_error)
						else:
							click.secho(line_of_output.message, bold = line_of_output.is_error)

	if debug:
		time_elapsed = time.monotonic() - cq_started
		logging.debug('cq ended, time elapsed: %.3f s', time_elapsed)

		logger.debug('slowest runners:')

		module_time_elapsed.sort(key = lambda x: x.time, reverse = True)

		for run in module_time_elapsed[:3]:
			logger.debug('%s with %.3f s', run.name, run.time)
	return exit_code


def run_fixers(
	list_modules: List[str],
	disable: Iterable[str],
	debug: bool,
) -> int:
	if debug:
		logging.debug('cq started fixing in debug mode')
	exit_code = 0
	return_codes: List[int] = []

	for fixer in fixers:
		if fixer.NAME in disable:
			if debug:
				logging.debug('module %s disabled -> skipping', fixer.NAME)
			continue
		click.secho(fixer.NAME, bg = 'white', fg = 'black', bold = True, underline = True)
		return_codes.append(fixer.run(list_modules))
	exit_code = max(return_codes)
	return exit_code


@click.command()
@click.option('--debug', help = 'Display debbuging information.', is_flag = True)
@click.option('--fix', help = 'Run fixers instead of checkers', is_flag = True)
@click.option('--show-warnings', help = 'Display warnings.', is_flag = True)
@click.option('-j', default = None, type = int, help = 'Number of threads to use.')
@click.option(
	'--disable',
	'-d',
	multiple = True,
	help = 'Disable one of the checkers, e.g. -d pyflakes-ext -d mypy.',
)
@click.option(
	'--help',
	'-h',
	callback = help_callback,
	is_flag = True,
	expose_value = False,
	is_eager = True,
	help = 'Show this message and exit.',
)
@click.argument('modules', nargs = -1)
def main(
	debug: bool,
	fix: bool,
	show_warnings: bool,
	j: Optional[int],
	disable: Tuple[str],
	modules: Tuple[str],
) -> None:
	logging.basicConfig(
		format = '%(asctime)s: %(levelname)s: %(message)s',
		handlers = [
			logging.StreamHandler(),
		],
		level = logging.DEBUG if debug else logging.INFO,
	)

	logger = logging.getLogger(__name__)

	if not modules:
		list_modules = ['.']
	else:
		list_modules = list(modules)

	result_char = CHECK_MARK
	result_color = 'green'

	if fix:
		exit_code = run_fixers(list_modules, disable, debug,)
	else:
		exit_code = run_checkers(list_modules, disable, logger, debug, show_warnings, j)

	if exit_code > 0:
		result_char = FAIL_X
		result_color = 'red'

	click.secho(result_char, fg = result_color)
	sys.exit(exit_code)


if __name__ == '__main__':
	main()  # pylint: disable = no-value-for-parameter



from typing import List
import contextlib
import os

import isort

import cq.checker
import cq.checkers.isort_config
import cq.utils


class IsortChecker(cq.checker.Checker):
	NAME = 'isort'
	HELP_LINE = 'Run `cq --fix`'
	DESCRIPTION = 'Checks whether imports are sorted according to the code style.'

	def run(self, modules: List[str]) -> cq.checker.CheckerResult:
		# This redirects all stderr to /dev/null.
		# Because of multithreading, all checkers running in parallel will have stderr redirected as well.
		# This doesn't cause problems now (while checkers use stdout) but might in the future.
		with open(os.devnull, 'w') as devnull, contextlib.redirect_stderr(devnull):
			results = {
				python_file: isort.check_file(python_file, config = cq.checkers.isort_config.Config)
				for python_file in cq.utils.get_all_python_files(modules)
			}

		return cq.checker.CheckerResult(
			checker_name = self.NAME,
			help_line = self.HELP_LINE,
			return_code = 0,
			output = [
				cq.checker.ResultLine(
					file = filename,
					line = None,
					message = 'Imports are incorrectly sorted and/or formatted.',
					is_error = False,
				)
				for filename, result in results.items()
				if result is False
			],
		)



from typing import List, Set
import os.path

import bellybutton.cli
import bellybutton.parsing

import cq.checker
import cq.utils


class BellybuttonChecker(cq.checker.Checker):
	NAME = 'bellybutton'
	DESCRIPTION = 'Launch Bellybutton linter on given modules.'
	HELP_LINE = 'use `# bb: ignore` for disabling bellybutton on particular line'

	def run(self, modules: List[str]) -> cq.checker.CheckerResult:
		files: Set[str] = set()

		for file in cq.utils.get_all_python_files(modules):
			files.add(file)

		bellybutton_conf = os.path.join(os.path.abspath(os.path.dirname(__file__)), '.bellybutton.yml')
		with open(bellybutton_conf, 'r') as f:
			rules = bellybutton.parsing.load_config(f)

		results = bellybutton.cli.linting_failures(files, rules)

		output: List[cq.checker.ResultLine] = []
		for result in results:
			output.append(
				cq.checker.ResultLine(
					file = result.path,
					line = result.lineno,
					message = result.rule.description,
					is_error = True,
				)
			)

		return cq.checker.CheckerResult(
			checker_name = self.NAME,
			help_line = self.HELP_LINE,
			return_code = 0 if not output else 1,
			output = output,
		)



from typing import List, NamedTuple, Optional
import abc


class ResultLine(NamedTuple):
	file: Optional[str]
	line: Optional[int]
	message: str
	is_error: bool


class CheckerResult(NamedTuple):
	checker_name: str
	help_line: Optional[str]
	return_code: int
	output: List[ResultLine]


class Checker(abc.ABC):
	NAME: str
	DESCRIPTION: str

	@abc.abstractmethod
	def run(self, modules: List[str]) -> CheckerResult:
		pass
